#############
# MIT License
#
# Copyright (C) 2020 ThoughtForge Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy 
# of this software and associated documentation files (the "Software"), to deal 
# in the Software without restriction, including without limitation the rights 
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
# copies of the Software, and to permit persons to whom the Software is 
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all 
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
# SOFTWARE.
#############

import gym
import math 
import os
import numpy as np

from thoughtforge_client.thoughtforge_client import BaseThoughtForgeClientSession


##################################################################
### Modify Reacher gym to not reset its arm position
from gym.envs.mujoco.reacher import ReacherEnv

last_qpos = None
last_qvel = None

def new_reset(self):
    global last_qpos, last_qvel
    last_qpos = np.copy(self.sim.data.qpos)
    last_qvel = np.copy(self.sim.data.qvel)
    self.sim.reset()
    ob = self.reset_model()
    return ob

ReacherEnv.reset = new_reset

def new_reset_model(self):
    # qpos = self.np_random.uniform(low=-0.1, high=0.1, size=self.model.nq) + self.init_qpos
    qpos = np.copy(last_qpos)
    qvel = np.copy(last_qvel)
    while True:
        self.goal = self.np_random.uniform(low=-.2, high=.2, size=2)
        if np.linalg.norm(self.goal) < 0.2:
            break
    qpos[-2:] = self.goal
    # qvel = self.init_qvel + self.np_random.uniform(low=-.005, high=.005, size=self.model.nv)
    qvel[-2:] = 0
    self.set_state(qpos, qvel)
    return self._get_obs()

ReacherEnv.reset_model = new_reset_model
##################################################################


gym.register(
    id='reacher_3joint-v0',
    entry_point='advanced.reacher_3joint.reacher_3joint_env:Reacher3JointEnv',
    max_episode_steps=500,
)

def angle_between(v1, v2):
    def unit_vector(vector):
        return vector / np.linalg.norm(vector)
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))



class ExampleReacher3Session(BaseThoughtForgeClientSession):

    def _reset_env(self):
        """ local helper function specific for openAI gym environments """
        self.last_observation = self.env.reset()
        self.score = 0
        # for tracking first couple derivatives of motion
        self.last_rotvel0 = 0
        self.last_rotvel1 = 0
        self.last_delta1_rotvel0 = 0
        self.last_delta1_rotvel1 = 0

    def sim_started_notification(self):
        """ On sim start, initialize environment """   
        print("Initializing reacher_3joint-v0...", end='')
        self.env = gym.make('reacher_3joint-v0')
        if self.env is not None:
            self._reset_env()
        print("Complete.")

    def sim_ended_notification(self):
        """ On sim start, destroy environment """
        if self.env is not None:
            self.env.close()
            self.env = None
        
    def update(self, motor_dict):
        """ advance the environment sim """
        # render the environment locally
        self.env.render()

        # extract action sent from server
        motor_value_0 = motor_dict['motor_value_0'][0]
        motor_value_1 = motor_dict['motor_value_1'][0]
        motor_value_2 = motor_dict['motor_value_2'][0]
        
        # step openAI gym env and get updated observation
        env_step_result = self.env.step([motor_value_0, motor_value_1, motor_value_2])
        self.last_observation, _reward_unused, terminal, _ = env_step_result
        self.last_observation, reward, terminal, _ = env_step_result
        self.score += reward
        if terminal:
            print("End of episode. Score =", self.score)
            self._reset_env()
 
        # send updated environment data to server
        body0_pos = self.env.get_body_com("body0")
        body1_pos =self.env.get_body_com("body1")
        fingertip_pos = self.env.get_body_com("fingertip")
        target_pos = self.env.get_body_com("target")

        body0_to_target = target_pos - body0_pos
        body0_to_fingertip = fingertip_pos - body0_pos
        angle0_sign = np.sign(np.cross(body0_to_target, body0_to_fingertip)[2])
        angle0 = angle_between(body0_to_target, body0_to_fingertip)/np.pi
        # angle0_sensor
        angle0_sensor_val = angle0_sign * angle0
        
        clamped_abs_angle0 = angle0
        if angle0 < 0.25:
            clamped_abs_angle0 = 0.25        
        body0_xvelr = self.env.data.get_body_xvelr("body0")[2]
        body0_xvelr_scaled = body0_xvelr/10
        if abs(body0_xvelr_scaled) < 0.04:
            body0_xvelr_scaled = 0
        # rotvel0_sensor
        rotvel0_sensor_val = (body0_xvelr_scaled) / clamped_abs_angle0

        new_delta1_rotvel0 = self.last_rotvel0 - body0_xvelr
        # delta1_rotvel0_sensor
        delta1_rotvel0_sensor_val = (new_delta1_rotvel0 / clamped_abs_angle0)
        self.last_rotvel0 = body0_xvelr

        new_delta2_rotvel0 = self.last_delta1_rotvel0 - new_delta1_rotvel0
        # delta2_rotvel0_sensor
        delta2_rotvel0_sensor_val = (new_delta2_rotvel0  / clamped_abs_angle0)
        self.last_delta1_rotvel0 = new_delta1_rotvel0

        # radius_sensor
        radius_sensor_val = np.linalg.norm(body0_to_target) - np.linalg.norm(body0_to_fingertip)

        body1_xvelr = self.env.data.get_body_xvelr("body1")[2]
        # rotvel1_sensor
        rotvel1_sensor_val = body0_xvelr/20
        if abs(rotvel1_sensor_val) < 0.08:
            rotvel1_sensor_val = 0

        # delta1_rotvel1_sensor
        delta1_rotvel1_sensor_val = self.last_rotvel1 - body1_xvelr
        self.last_rotvel1 = body1_xvelr

        # delta2_rotvel1_sensor
        delta2_rotvel1_sensor_val = self.last_delta1_rotvel1 - delta1_rotvel1_sensor_val
        self.last_delta1_rotvel1 = delta1_rotvel1_sensor_val
        
        sensor_values = {
            "angle0_sensor1" : angle0_sensor_val,
            "angle0_sensor2" : angle0_sensor_val,
            "angle0_sensor3" : angle0_sensor_val,
            "angle0_sensor4" : angle0_sensor_val,
            'rotvel1_sensor': rotvel1_sensor_val,
            'delta1_rotvel1_sensor': delta1_rotvel1_sensor_val,
            'delta2_rotvel1_sensor': delta2_rotvel1_sensor_val,
            'radius_sensor1': radius_sensor_val,
            'radius_sensor2': radius_sensor_val,
            'rotvel0_sensor': rotvel0_sensor_val,
            'delta1_rotvel0_sensor': delta1_rotvel0_sensor_val,
            'delta2_rotvel0_sensor': delta2_rotvel0_sensor_val
        }
        return sensor_values


if __name__ == "__main__": 
    session = ExampleReacher3Session.from_file('./advanced/reacher_3joint/example_reacher3.params')
